from .experiment import Experiment
from mixed_type_diffusion import MixedTypeDiffusion
from layers.mlp import MLP, TabDDPM_MLP
from layers.improved_cdtd import MPMLP, log_score_model_weights
from train_utils import cycle
import torch
from tqdm.autonotebook import tqdm
import plotting
import logging
from torch.utils.tensorboard import SummaryWriter
import os
from pprint import pformat
import numpy as np
from torch_ema import ExponentialMovingAverage
from utils import get_total_trainable_params, set_seeds
import time
from train_utils import LinearScheduler, InverseSquareRootScheduler


class Experiment_CDTD(Experiment):
    """Train and evaluate a continuous diffusion model for mixed-type tabular data."""
    
    def __init__(self, config, exp_path, dataset):
        super().__init__(config, exp_path, dataset)
    
    
    def get_model(self):
        
        args = self.config.model
        self.categories = self.data_wrangler.num_cats
        self.num_cat_features = self.data_wrangler.num_cat_features
        self.num_cont_features = self.data_wrangler.num_cont_features
        self.num_features = self.data_wrangler.num_total_features
        mp_cat = False
        self.simulate_missings = self.config.training.get('simulate_missings', False)
        
        # derive proportions for max CE losses at t = 1 for normalization
        self.normalize_losses = args.normalize_losses
        self.adaptive_weight = args.adaptive_weight

        if self.normalize_losses:
            proportions = []
            n_sample = self.train_loader.X_cat.shape[0]
            for i in range(len(self.categories)):
                _, counts = self.train_loader.X_cat[:,i].unique(return_counts=True)
                proportions.append(counts / n_sample)
            self.proportions = proportions
        else:
            self.proportions = None
        
        # relative weights in case of no loss normalization
        loss_weights = torch.cat((torch.tensor(args.ce_weight).repeat((1, self.num_cat_features)),
                                      torch.tensor((1-args.ce_weight)).repeat((1, self.num_cont_features))), 
                                     dim = 1)
        self.rel_loss_weights = (loss_weights / loss_weights.sum()).to(self.device)
                 
        if self.config.model.architecture == 'mlp':
            score_model = MLP(self.num_cont_features, args.dim, 
                              self.categories, self.data_wrangler.num_y_classes,
                              args.mlp_emb_dim, args.mlp_n_layers, args.mlp_n_units,
                              proportions=self.proportions, use_fourier_features=args.use_fourier_features,
                              act=args.act, feat_spec_cond=args.use_feat_spec_cond, time_fourier=args.use_time_fourier)
            
        elif self.config.model.architecture == 'mpmlp':
            mp_cat = True
            score_model = MPMLP(args.mlp_emb_dim, args.dim, self.num_cont_features,
                                self.categories, self.data_wrangler.num_y_classes, args.mlp_n_layers,
                                args.mlp_n_units, proportions=self.proportions, feat_spec_cond=args.use_feat_spec_cond)
            
        elif self.config.model.architecture == 'tabddpm':
            score_model = TabDDPM_MLP(self.num_cont_features, args.dim, 
                              self.categories, self.data_wrangler.num_y_classes,
                              args.mlp_emb_dim, args.mlp_n_layers, args.mlp_n_units,
                              proportions=self.proportions, use_fourier_features=args.use_fourier_features)
        
        return MixedTypeDiffusion(model = score_model, 
                                  dim = args.dim,
                                  categories = self.categories,
                                  num_features = self.num_features,
                                  task = self.data_wrangler.task,
                                  sigma_data_cat = args.sigma_data_cat, 
                                  sigma_data_cont = args.sigma_data_cont,
                                  sigma_min_cat = args.sigma_min_cat,
                                  sigma_max_cat = args.sigma_max_cat,
                                  sigma_min_cont = args.sigma_min_cont,
                                  sigma_max_cont = args.sigma_max_cont,
                                  normalize_losses = args.normalize_losses,
                                  proportions = self.proportions,
                                  cat_emb_init_sigma = args.cat_emb_init_sigma,
                                  timewarp_variant = args.timewarp_variant,
                                  timewarp_type = args.timewarp_type,
                                  timewarp_weight_low_noise = args.timewarp_weight_low_noise,
                                  timewarp_bins = args.timewarp_bins,
                                  timewarp_decay = args.timewarp_decay,
                                  adaptive_weight = args.adaptive_weight,
                                  cat_bias=args.use_cat_bias,
                                  mp_cat=mp_cat,
                                  simulate_missings=self.simulate_missings
                                  )
            
            
    def get_optimizer(self):    
        config = self.config.optimizer

        if config.name == 'adam':
            optimizer = torch.optim.Adam(self.diff_model.parameters(), **config.args)
            if hasattr(config, 'gradient_clip_norm'):
                NotImplementedError()
        elif config.name == 'adamw':
            optimizer = torch.optim.AdamW(self.diff_model.parameters(), **config.args)
        else:
            raise Exception('Unknown optimizer.')

        return optimizer


    def get_lr_scheduler(self):
        
        if self.config.training.scheduler == 'linear':
            
            scheduler = LinearScheduler(self.config.training.num_steps_train,
                                        base_lr = self.config.optimizer.args.lr,
                                        final_lr = 1e-6,
                                        warmup_steps = self.config.training.num_steps_lr_warmup,
                                        warmup_begin_lr = 1e-6,
                                        anneal_lr = self.config.training.anneal_lr)
        else:
            
            scheduler = InverseSquareRootScheduler(base_lr = self.config.optimizer.args.lr, 
                                                ref_step = self.config.training.ref_step, 
                                                warmup_steps = self.config.training.num_steps_lr_warmup, 
                                                anneal_lr=self.config.training.anneal_lr, 
                                                warmup_begin_lr=1e-6)
        return scheduler    
        
        
    def train(self, **kwargs):
        
        plot_figures = kwargs.get('plot_figures', False)
        save_model = kwargs.get('save_model', False)
        eval_model = kwargs.get('eval_model', False)
        
        # get train data
        self.train_loader = self.data_wrangler.get_train_loader(self.config.training.batch_size)
        train_iter = cycle(self.train_loader)
    
        # initialize model
        logging.warning('=== Initializing model ===')
        set_seeds(self.seed, cuda_deterministic=True)
        self.diff_model = self.get_model().to(self.device)
        print(f'total number of trainable parameters = {get_total_trainable_params(self.diff_model)}.')
        self.ema_diff_model = ExponentialMovingAverage(self.diff_model.parameters(), decay = self.config.optimizer.ema_decay)
        
        # initialize optimizer
        self.optimizer = self.get_optimizer()
        self.scheduler = self.get_lr_scheduler()
        
        if plot_figures:
            self.evaluate_loss_over_time(self.train_loader, prefix='init_')
        
        logging.warning('=== Start training... ===')
        config = self.config.training
  
        # initialize logging
        self.current_step = 0
        if save_model:
            writer = SummaryWriter(os.path.join(self.logdir, 'tb'))
            with open(os.path.join(self.logdir, 'hparams.txt'), 'w') as f:
                    f.write(pformat(self.config))
                    
        gif_frames = {'timewarp_cdf': [], 'diff_in_means': [], 'diff_in_corr': []}
        training_start_time = time.time()
        
        with tqdm(initial=self.current_step, total=config.num_steps_train, disable=(not save_model)) as pbar:
            while self.current_step < config.num_steps_train:
                
                is_last_step = self.current_step == (config.num_steps_train - 1)
                self.optimizer.zero_grad()
                       
                inputs = next(train_iter)
                x_cat, x_cont, y_cond = (input.to(self.device) if input is not None else None for input in inputs)
                
                unweighted_losses, weighted_losses, weightnet_losses, sigmas = self.diff_model.loss_fn(x_cat, x_cont, y_cond, return_sigma=True)
                assert unweighted_losses.shape == sigmas.shape == (x_cat.shape[0], self.num_features)
                cdf_losses = self.diff_model.timewarp_cdf.loss_fn(sigmas.detach(), unweighted_losses.detach())
                
                if not self.normalize_losses:
                    # weighted average of CE and MSE losses with pre-specified weight
                    diff_loss = (weighted_losses * self.rel_loss_weights).sum(1)
                else:
                    diff_loss = weighted_losses.mean(1)
                
                diff_loss.mean().backward()
                cdf_losses.mean().backward()
                if self.adaptive_weight:
                    weightnet_losses.mean().backward()
                
                # update parameters
                self.optimizer.step()
                self.diff_model.timewarp_cdf.update_ema()
                self.ema_diff_model.update()
                self.current_step += 1
                pbar.update(1)
                
                # anneal learning rate
                if self.scheduler:
                    for param_group in self.optimizer.param_groups:
                        param_group["lr"] = self.scheduler(self.current_step)
             
                # track metric
                train_dict, train_dict_feat_spec = self.get_metric_dict(diff_loss, weighted_losses, unweighted_losses, cdf_losses, weightnet_losses)
    
                # plot histograms of average generated sigmas (slows down training a lot!)
                if save_model and plot_figures:
                    writer.add_histogram('average sigma over features', sigmas.mean(1), self.current_step)
                    writer.add_histogram('average sigma over cat features', sigmas[:, :self.num_cat_features].mean(1), self.current_step)
                    writer.add_histogram('average sigma over cont features', sigmas[:, self.num_cat_features:].mean(1), self.current_step)
                    writer.add_histogram('average log sigma over cat features', sigmas[:, :self.num_cat_features].log().mean(1), self.current_step)
                    writer.add_histogram('average log sigma over cont features', sigmas[:, self.num_cat_features:].log().mean(1), self.current_step)
                
                # evaluation loop
                eval_dict = eval_dict_feat_spec = None
        
                if eval_model:
                    eval_dict = {}
                    eval_dict_feat_spec = {}
                    if self.current_step % config.steps_per_eval == 0 or is_last_step:
                        with self.ema_diff_model.average_parameters():
                            
                            X_cat_gen, X_cont_gen, y_gen = self.sample_tabular_data(1000, verbose=False)
                
                            # compute diff in means
                            diff_in_means = self.data_wrangler.compute_diff_in_means(X_cat_gen, X_cont_gen, y_gen)
                            eval_dict_feat_spec['diff_in_means'] = diff_in_means
                            eval_dict['mean_mse_diff'] = (diff_in_means ** 2).mean()
                            
                            # save figure of absolute diff in means
                            gif_frames['diff_in_means'].append(plotting.plot_diff_in_means_gif_frame(diff_in_means.abs(), self.current_step))
                            
                            # compute L2 norm between correlation matrices
                            diff_in_corr = self.data_wrangler.compute_diff_in_corr(X_cat_gen, X_cont_gen, y_gen)
                            eval_dict['diff_in_corr_l2norm'] = diff_in_corr['l2_norm']
                            
                            # save figure of absolute diff in correlation matrices
                            gif_frames['diff_in_corr'].append(plotting.plot_diff_in_corr_gif_frame(diff_in_corr['abs'], self.current_step))
                    
                if plot_figures:
                    if self.current_step % config.steps_per_eval == 0 or is_last_step:
                        with self.ema_diff_model.average_parameters():
                            # save figure of timewarp cdfs
                            gif_frames['timewarp_cdf'].append(plotting.plot_loss_gif_frame(self.diff_model))
                                
                # write to summaryfile
                if save_model:
                    if self.current_step % config.steps_per_logging == 0 or self.current_step % config.steps_per_eval == 0 or is_last_step:
                        # log_score_model_weights(writer, self.current_step, self.diff_model)                   
                        self.log_fn(writer, self.current_step, train_dict, eval_dict)
                        # self.feature_specific_log_fn(writer, train_dict_feat_spec, eval_dict_feat_spec)
                        
        # compute training duration
        training_duration = time.time() - training_start_time
        self.save_train_time(training_duration)
        
        # take EMA of model parameters
        self.ema_diff_model.copy_to() 
        self.diff_model.eval()
            
        # generate timewarping cdf gif
        if plot_figures: 
            self.plot_figures(gif_frames, self.train_loader)
        
        if save_model: 
            self.save_model()
            writer.close()
        logging.info('=== Finished model training. ===')  
    
        
    def evaluate_loss_over_time(self, data_loader, steps = 100, prefix = 'train_'):
        """Generate figures of loss over time over whole set.

        Args:
            split (str): train, val, test set selection
            steps (int, optional): number of timesteps to form the grid. Defaults to 100.
        """
        
        u_list = torch.linspace(1e-7, 1-1e-7, steps, device=self.device)
        X_cat = data_loader.X_cat.clone()
        X_cont = data_loader.X_cont.clone()
        y = data_loader.y.clone() if self.config.model.y_cond else None
        
        # if full dataset is too large, subsample 50k
        train_obs = self.data_wrangler.data.get_train_obs()
        if train_obs > 50000:
            idx = torch.randperm(X_cat.shape[0], device=X_cat.device)[:50000]
            X_cat = X_cat[idx]
            X_cont = X_cont[idx]
            if y is not None:
                y = y[idx]
            train_obs = 50000
        
        data_split = torch.split(torch.arange(train_obs), 3000)

        data_dict = {'weighted_avg': [],
                     'weighted_std': [],
                     'unweighted_avg': [],
                     'unweighted_std': [],
                     'cdf_avg': [],
                     'cdf_std': [],
                     'weightnet_avg': [],
                     'weightnet_std': [],
                     'sigma': []}

        for u in tqdm(u_list):
            
            weighted_losses_per_u = []
            unweighted_losses_per_u = []
            cdf_losses_per_u = []
            weightnet_losses_per_u = []
            sigmas = None
            
            for idx in data_split:
                
                x_cat = X_cat[idx].to(self.device)
                x_cont = X_cont[idx].to(self.device)
                
                if self.config.model.y_cond:
                    y_cond = y[idx].to(self.device)
                else:
                    y_cond = None
                
                batch = x_cat.shape[0] if x_cat is not None else x_cont.shape[0]
                u_tensor = torch.full((batch,), fill_value=u, device=self.device)
                with torch.no_grad():
                    # note that this randomly includes/excludes self-conditioning!
                    unweighted_losses, weighted_losses, weightnet_losses, sigmas = self.diff_model.loss_fn(x_cat, x_cont, y_cond, return_sigma=True, u=u_tensor)
                   
                    cdf_losses = self.diff_model.timewarp_cdf.loss_fn(sigmas.detach(), unweighted_losses.detach())
                    
                weighted_losses_per_u.append(weighted_losses.cpu())
                unweighted_losses_per_u.append(unweighted_losses.cpu())
                cdf_losses_per_u.append(cdf_losses.cpu())
                if weightnet_losses is not None:
                    weightnet_losses_per_u.append(weightnet_losses.cpu().squeeze())
                
            data_dict['sigma'].append(sigmas[0,:].cpu())
            
            # average over batch dimension
            weighted_loss = torch.cat(weighted_losses_per_u)
            data_dict['weighted_avg'].append(weighted_loss.mean(0))
            data_dict['weighted_std'].append(weighted_loss.std(0))
            
            unweighted_loss = torch.cat(unweighted_losses_per_u)
            data_dict['unweighted_avg'].append(unweighted_loss.mean(0))
            data_dict['unweighted_std'].append(unweighted_loss.std(0))
            
            cdf_loss = torch.cat(cdf_losses_per_u)
            data_dict['cdf_avg'].append(cdf_loss.mean(0))
            data_dict['cdf_std'].append(cdf_loss.std(0))
            
            weightnet_loss = torch.cat(weightnet_losses_per_u) if weightnet_losses is not None else None
            if weightnet_loss is not None:
                data_dict['weightnet_avg'].append(weightnet_loss.mean(0))
                data_dict['weightnet_std'].append(weightnet_loss.mean(0))
            else:
                data_dict['weightnet_avg'] = None
                data_dict['weightnet_std'] = None
            
        # construct tensors from lists
        data_dict = {k: torch.stack(v) for k,v in data_dict.items() if v is not None}
        data_dict['u'] = u_list.detach().cpu()
        
        plotting.plot_cdf_and_loss(self.logdir, self.diff_model, 
                                   data_dict, self.num_cat_features,
                                   prefix=prefix)
            
        plotting.plot_weighting(self.logdir, self.diff_model, prefix=prefix)


        # PLOT WEIGHTED LOSSES
        if self.num_cat_features > 0:
            ce_loss = data_dict['weighted_avg'][:, :self.num_cat_features]
            sigma_cat = data_dict['sigma'][:, :self.num_cat_features]
            max_ce = plotting.get_max_ce(X_cat, y, self.categories,)
            
            plotting.plot_cat_losses_wrt_sigma(self.logdir, self.categories, sigma_cat, ce_loss, max_ce, prefix = prefix + 'weighted_', scale_x = False)
            
            plotting.plot_cat_losses_wrt_sigma(self.logdir, self.categories, sigma_cat, ce_loss, max_ce, prefix = prefix + 'weighted_', scale_x = True)
            
            plotting.plot_cat_losses_wrt_u(self.logdir, self.categories, data_dict['u'], ce_loss, prefix = prefix + 'weighted_')
            

        if self.num_cont_features > 0:
            mse_loss = data_dict['weighted_avg'][:, self.num_cat_features:]
            sigma_cont = data_dict['sigma'][:, self.num_cat_features:]
   
            plotting.plot_cont_losses_wrt_sigma(self.logdir, sigma_cont, mse_loss, prefix = prefix + 'weighted_', scale_x = False)
            
            plotting.plot_cont_losses_wrt_sigma(self.logdir, sigma_cont, mse_loss, prefix = prefix + 'weighted_', scale_x = True)
            
            plotting.plot_cont_losses_wrt_u(self.logdir, data_dict['u'], mse_loss, prefix = prefix + 'weighted_')
            
        
        # PLOT UNWEIGHTED LOSSES
        if self.num_cat_features > 0:
            ce_loss = data_dict['unweighted_avg'][:, :self.num_cat_features]
            ce_loss_std = data_dict['unweighted_std'][:, :self.num_cat_features]
            sigma_cat = data_dict['sigma'][:, :self.num_cat_features]
            max_ce = plotting.get_max_ce(X_cat, y, self.categories,)
            
            plotting.plot_cat_losses_wrt_sigma(self.logdir, self.categories, sigma_cat, ce_loss, max_ce, prefix = prefix + 'unweighted_', scale_x = False)
            
            plotting.plot_cat_losses_wrt_sigma(self.logdir, self.categories, sigma_cat, ce_loss, max_ce, prefix = prefix + 'unweighted_', scale_x = True)
            
            plotting.plot_cat_losses_wrt_u(self.logdir, self.categories, data_dict['u'], ce_loss, prefix = prefix + 'unweighted_')
        
        
        if self.num_cont_features > 0:
            mse_loss = data_dict['unweighted_avg'][:, self.num_cat_features:]
            mse_loss_std = data_dict['unweighted_std'][:, self.num_cat_features:]
            sigma_cont = data_dict['sigma'][:, self.num_cat_features:]
   
            plotting.plot_cont_losses_wrt_sigma(self.logdir, sigma_cont, mse_loss, prefix = prefix + 'unweighted_', scale_x = False)
            
            plotting.plot_cont_losses_wrt_sigma(self.logdir, sigma_cont, mse_loss, prefix = prefix + 'unweighted_', scale_x = True)
            
            plotting.plot_cont_losses_wrt_u(self.logdir, data_dict['u'], mse_loss, prefix = prefix + 'unweighted_')
            
            
        # PLOT AVERAGE LOSSES in uniform time
        plotting.plot_avg_losses_wrt_u(self.logdir, prefix, data_dict, self.num_cat_features)
        
        
        # PLOT TIMEWARPING LOSS APPROXIMATION WRT TRUE DIFF LOSS
        plotting.plot_cdf_losses_wrt_sigma(self.logdir, self.diff_model.timewarp_type,
                                           data_dict['sigma'], data_dict['cdf_avg'], self.num_cat_features, prefix)
            
        
        
    def sample_tabular_data(self, num_samples, **kwargs):
        
        seed = kwargs.get('seed', None)
        verbose = kwargs.get('verbose', False)
        
        if seed:
            set_seeds(seed, cuda_deterministic=True)
        
        if self.config.model.y_cond:
            _, _, y_train = self.data_wrangler.data.get_train_data()
            _, y_counts = np.unique(y_train, return_counts = True)
            y_dist = (y_counts / y_counts.sum())
        else:
            y_dist = None
    
        X_cat_gen, X_cont_gen, y_gen = self.diff_model.sample_data(num_samples, 
                                                                   self.config.training.batch_size_eval, 
                                                                   y_dist, 
                                                                   self.config.model.generation_steps,
                                                                   verbose=verbose)
        
        X_cat_gen, X_cont_gen, y_gen = self.data_wrangler.postprocess_gen_data(X_cat_gen.to(torch.long).numpy(), X_cont_gen.numpy(), y_gen.numpy() if y_gen is not None else None)
        X_cat_gen = X_cat_gen.astype(int)
        
        return X_cat_gen, X_cont_gen, y_gen
    
    
    def impute_data(self, X_cat, X_cont, y, mask, num_steps, resample_steps, **kwargs):
        
        seed = kwargs.get('seed', None)
        verbose = kwargs.get('verbose', False)
        
        if seed:
            set_seeds(seed, cuda_deterministic=True)
            
        X_cat_gen, X_cont_gen, y_gen = self.diff_model.impute_data(X_cat, 
                                                                   X_cont, 
                                                                   y,
                                                                   mask, 
                                                                   self.config.training.batch_size_eval, 
                                                                   num_steps,
                                                                   resample_steps=resample_steps,
                                                                   verbose=verbose)
        
        # X_cat_gen, X_cont_gen, y_gen = self.data_wrangler.postprocess_gen_data(X_cat_gen.to(torch.long).numpy(), X_cont_gen.numpy(), y_gen.numpy() if y_gen is not None else None)
        X_cat_gen = X_cat_gen.numpy().astype(int)
        X_cont_gen = X_cont_gen.numpy()
        y_gen = y_gen.numpy() if y_gen is not None else None
        
        return X_cat_gen,  X_cont_gen, y_gen
            

    def get_metric_dict(self, diff_loss, losses, unweighted_losses, cdf_losses, weighnet_loss):
                    
        if weighnet_loss is not None:
            weighnet_loss = weighnet_loss.mean()
            
        total_loss = diff_loss.mean()
                    
        # metrics for main model
        # rescale_to_bpd = 1. / (self.num_features * torch.tensor(2.).log())
        feat_spec_loss = losses.detach().mean(0)
        loss_avg_over_cat_feat = losses[:, :self.num_cat_features].detach().mean(1).mean(0).item() if self.num_cat_features > 0 else 0
        loss_avg_over_cont_feat = losses[:, self.num_cat_features:].mean(1).detach().mean(0).item() if self.num_cont_features > 0 else 0
        avg_loss_prop = (losses / losses.sum(1).unsqueeze(1)).mean(0).detach() 

        # metrics for timewarping
        individual_timewarp_loss = cdf_losses.detach().mean(0)
        total_timewarp_loss = cdf_losses.detach().mean().item()
        
        unweight_feat_spec_loss = unweighted_losses.detach().mean(0)
        unweight_loss_avg_over_cat_feat = unweighted_losses[:, :self.num_cat_features].detach().mean(1).mean(0).item() if self.num_cat_features > 0 else 0
        unweight_loss_avg_over_cont_feat = unweighted_losses[:, self.num_cat_features:].mean(1).detach().mean(0).item() if self.num_cont_features > 0 else 0
        
        scalar_dict = {
            'total_weighted_loss': total_loss.detach().item(),
            'timewarp/total_loss': total_timewarp_loss,
        }
        
        if weighnet_loss is not None:
            scalar_dict['weightnet_loss'] = weighnet_loss.detach().item()
        
        feature_specific_dict = {
            'feat_specific_weighted_loss': feat_spec_loss,
            'weighted_loss_avg_over_cat_feat': loss_avg_over_cat_feat,
            'weighted_loss_avg_over_cont_feat': loss_avg_over_cont_feat,
            'feat_specific_unweighted_loss': unweight_feat_spec_loss,
            'unweighted_loss_avg_over_cat_feat': unweight_loss_avg_over_cat_feat,
            'unweighted_loss_avg_over_cont_feat': unweight_loss_avg_over_cont_feat,
            'timewarp/individual_loss': individual_timewarp_loss,
            'avg_weighted_loss_proportions': avg_loss_prop,
        }

        return scalar_dict, feature_specific_dict
    
    
    def log_fn(self, writer, step, train_dict, eval_dict):
        for metric_name, metric_value in train_dict.items():
            writer.add_scalar('train/{}'.format(metric_name), metric_value, global_step=step)
        if eval_dict:
            for metric_name, metric_value in eval_dict.items():
                writer.add_scalar('eval/{}'.format(metric_name), metric_value, global_step=step)
                
    
    def feature_specific_log_fn(self, writer, train_dict_feat_spec, eval_dict_feat_spec):
        """Logs feature-specific metrics to be displayed in a single graph."""
        
        labels = ['train', 'eval']
        for j, d in enumerate([train_dict_feat_spec, eval_dict_feat_spec]):
            
            if d is None:
                continue
        
            if j == 0:
                # handle training metrics
                feature_specific_losses = {f'feature{i}': d['feat_specific_weighted_loss'][i].item() for i in range(self.num_features)}
                feature_specific_unweighted_losses = {f'feature{i}': d['feat_specific_unweighted_loss'][i].item() for i in range(self.num_features)}
                timewarp_individual_losses = {f'{i}': d['timewarp/individual_loss'][i].item() for i in range(d['timewarp/individual_loss'].shape[0])}
                loss_proportions = {f'feature{i}': d['avg_weighted_loss_proportions'][i].item() for i in range(self.num_features)}
                
                bytype_losses = {'cat': d['weighted_loss_avg_over_cat_feat'],
                                'cont': d['weighted_loss_avg_over_cont_feat']}
                
                unweighted_bytype_losses = {'cat': d['unweighted_loss_avg_over_cat_feat'],
                                             'cont': d['unweighted_loss_avg_over_cont_feat']}
            
                writer.add_scalars(f'{labels[j]}/losses', feature_specific_losses, global_step=self.current_step)
                writer.add_scalars(f'{labels[j]}/unweighted_losses', feature_specific_unweighted_losses, global_step=self.current_step)
                writer.add_scalars(f'{labels[j]}/timewarp/ind_losses', timewarp_individual_losses, global_step=self.current_step)
                writer.add_scalars(f'{labels[j]}/bytype_losses', bytype_losses, global_step=self.current_step)
                writer.add_scalars(f'{labels[j]}/unweighted_bytype_losses', unweighted_bytype_losses, global_step=self.current_step)
                writer.add_scalars(f'{labels[j]}/weighted_avg_loss_proportions', loss_proportions, global_step=self.current_step)
                
            else:
                # handle eval metrics
                diff_in_means = {f'feature{i}': d['diff_in_means'][i].item() for i in range(self.num_features)}
                writer.add_scalars(f'{labels[j]}/diff_in_means', diff_in_means, global_step=self.current_step)


        
    def plot_figures(self, gif_frames, train_loader):
        
        if gif_frames['timewarp_cdf']:
            plotting.generate_timewarp_gif(self.logdir, gif_frames['timewarp_cdf'], uniform = False)
        if gif_frames['diff_in_means']:
            plotting.generate_diff_in_means_gif(self.logdir, gif_frames['diff_in_means'])
        if gif_frames['diff_in_corr']:
            plotting.generate_diff_in_corr_gif(self.logdir, gif_frames['diff_in_corr'])

        plotting.plot_timewarp_cdf(self.logdir, self.diff_model, scale_x=True, log_scale=False)
        plotting.plot_timewarp_pdf(self.logdir, self.diff_model, scale_x=True, log_scale=False)
        plotting.plot_timewarp_cdf(self.logdir, self.diff_model, scale_x=True, log_scale=True)
        plotting.plot_timewarp_pdf(self.logdir, self.diff_model, scale_x=True, log_scale=True)
        plotting.plot_timewarp_inverse_cdf(self.logdir, self.diff_model, scale_y=True)
        self.evaluate_loss_over_time(train_loader, prefix = 'train_')

        logging.info('=== Plots saved ===')
        
        
    def save_model(self):
        
        checkpoint = {'current_step': self.current_step,
                        'diff_model': self.diff_model._orig_mod.state_dict(),
                        'data_wrangler': self.data_wrangler,
                        'train_loader': self.train_loader}
        torch.save(checkpoint, os.path.join(self.ckpt_restore_dir, 'model.pt'))
        
        
    def load_model(self):

        checkpoint = torch.load(os.path.join(self.ckpt_restore_dir, 'model.pt'))
        self.current_step = checkpoint['current_step']
        self.train_loader = checkpoint['train_loader']
        self.data_wrangler = checkpoint['data_wrangler']
        diff_model = self.get_model()
        diff_model.load_state_dict(checkpoint['diff_model'])
        self.diff_model = torch.compile(diff_model)
        self.diff_model.to(self.device)
        self.diff_model.eval()
        
        
    def suggest_params(self, trial):
        # self.config.model.dim = 2 ** trial.suggest_int("emb_dim", 6, 7)
        # self.config.model.depth = trial.suggest_int("depth", 2, 4)
        # self.config.model.heads = trial.suggest_categorical("heads", [4, 8, 16])
        # self.config.model.cat_emb_init_sigma = trial.suggest_categorical("cat_emb_init_sigma", [0.001, 0.01, 0.1])
        # self.config.optimizer.args.lr = trial.suggest_float("lr", 0.0001, 0.002, log = True)
        # self.config.optimizer.args.weight_decay = trial.suggest_categorical("weight_decay", [0.001, 0])
        # self.config.training.num_steps_train = trial.suggest_categorical("num_steps_train", [10000, 15000, 20000])
        
        # if not self.config.model.normalize_losses:
        #     self.config.model.ce_weight = trial.suggest_categorical("ce_weight", [0.25, 0.4, 0.5, 0.6, 0.75])
            
        return self.config
