import torch
from torch import nn
import torch.nn.functional as F
from einops import rearrange, repeat
import numpy as np



class MPFourierEmbedding(nn.Module):
    def __init__(self, dim):
        super().__init__()
        assert (dim % 2) == 0
        half_dim = dim // 2
        self.weights = nn.Parameter(torch.randn(1, half_dim), requires_grad=False)
        self.register_buffer('scale', torch.tensor(2.).sqrt())

    def forward(self, x):
        freqs = x.unsqueeze(-1) * self.weights * 2 * torch.pi
        return torch.cat((freqs.sin(), freqs.cos()), dim = -1) * self.scale
    

class MPNumericalEmbedder(nn.Module):
    def __init__(self, dim, num_features, trainable=False):
        super().__init__()
        assert (dim % 2) == 0
        self.half_dim = dim // 2
        self.weights = nn.Parameter(torch.randn(1, num_features, self.half_dim), requires_grad=trainable)
        self.register_buffer('scale', torch.tensor(2.).sqrt())

    def forward(self, x):
        x = rearrange(x, 'b f -> b f 1')
        freqs = x * self.weights * 2 * torch.pi
        fourier = torch.cat((freqs.sin(), freqs.cos()), dim = -1) * self.scale
        return fourier
    

def normalize(x, eps=1e-4):
    """Karras paper Algorithm 1"""
    dim = list(range(1, x.ndim))
    n = torch.linalg.vector_norm(x, dim=dim, keepdim=True)
    alpha = np.sqrt(n.numel() / x.numel())
    return x / torch.add(eps, n, alpha=alpha)
    
    
class Linear(nn.Module):
    """
    Linear layer with forced weight normalization and without bias.
    Adapted from https://github.com/lucidrains/denoising-diffusion-pytorch/blob/main/denoising_diffusion_pytorch/karras_unet.py
    L159"""
    def __init__(self, dim_in, dim_out, concat_bias=False, init_zero=False):
        super().__init__()
        self.dim_in = dim_in + int(concat_bias)
        self.concat_bias = concat_bias
        self.weight = nn.Parameter(torch.randn(dim_out, self.dim_in))
        if init_zero:
            nn.init.zeros_(self.weight)
        
    def forward(self, x):
        
        if self.training:
            with torch.no_grad():
                self.weight.copy_(normalize(self.weight)) # var = 1
        weight = normalize(self.weight) / np.sqrt(self.dim_in) # norm = 1
        
        if self.concat_bias:
            x = torch.concat((x, torch.ones(*x.shape[:-1], 1, device=x.device)), dim=-1)
        
        return F.linear(x, weight)
    

class NLinear(nn.Module):
    def __init__(self, num_features, d_in, d_out, cond_dim = None, bias = False):
        super().__init__()
        
        self.num_features = num_features
        self.condition = (cond_dim is not None)
        self.layers = nn.ModuleList()
        for _ in range(num_features):
            layer = Linear(d_in, d_out, concat_bias=bias)
            self.layers.append(layer)
        self.act = MPSiLU()
        
        if self.condition:
            self.cond_proj = Linear(cond_dim, d_in * num_features)
            self.gain = nn.Parameter(torch.zeros(1, num_features, 1))

    def forward(self, x, c = None): 
        
        if self.condition:
            cond = self.cond_proj(c)
            cond = rearrange(cond, 'b (f d) -> b f d', f=self.num_features) * self.gain
            x *= cond  # * cond, gain = 0 initialized, so h = 0, and skip connection is zero at init
            
        h = torch.stack([l(x[:,i]) for i, l in enumerate(self.layers)], 1)
        return self.act(h)
    
    

class FeatCond(nn.Module):
    def __init__(self, num_features, d_in, d_out, cond_dim=None, bias=False):
        super().__init__()
        
        self.num_features = num_features
        self.condition = (cond_dim is not None)
        self.nlinear = MPNLinear(d_in, d_out, num_features, concat_bias=bias)
        self.act = MPSiLU()
        
        if self.condition:
            self.cond_proj = Linear(cond_dim, d_in * num_features)
            self.gain = nn.Parameter(torch.zeros(1, num_features, 1))

    def forward(self, x, c = None): 
        
        if self.condition:
            cond = self.cond_proj(c)
            cond = rearrange(cond, 'b (f d) -> b f d', f=self.num_features) * (1 + self.gain)
            x *= cond
        
        return self.act(self.nlinear(x))


class MPNLinear(nn.Module):
    """
    x has dimensions (batch, num_features, emb_dim)
    """
    def __init__(self, dim_in, dim_out, n, concat_bias=False):
        super().__init__()
        self.dim_in = dim_in + int(concat_bias)
        self.concat_bias = concat_bias
        self.weight = nn.Parameter(torch.randn(n, self.dim_in, dim_out))
        
    def forward(self, x):
        
        if self.training:
            with torch.no_grad():
                self.weight.copy_(self._normalize(self.weight))
        weight = self._normalize(self.weight) / np.sqrt(self.dim_in) # norm = 1
        
        if self.concat_bias:
            x = torch.concat((x, torch.ones(*x.shape[:-1], 1, device=x.device)), dim=-1)
    
        x = (x[..., None, :] @ weight).squeeze(-2)
        return x
    
    def _normalize(self, x, eps=1e-4):
        n = torch.linalg.vector_norm(x, dim=1, keepdim=True)
        alpha = np.sqrt(n[0].numel() / x[0].numel()) 
        return x / torch.add(eps, n, alpha=alpha)
        
        

class MPCatEmbedding(nn.Module):
    """
    Embedding corresponding to magnitude-preserving setup in Karras. Derives a one-hot vector from an index, scales it to unit variance, and then applies a magnitude-preserving linear layer. Allows for different number of categories per categorical feature and introduces a separate embedding for each category.
    """
    def __init__(self, emb_dim, categories, bias=False):
        super().__init__()
        self.categories = torch.tensor(categories)
        categories_offset = self.categories.cumsum(dim = -1)[:-1]
        categories_offset = torch.cat((torch.zeros((1,), dtype=torch.long), categories_offset))
        self.register_buffer('categories_offset', categories_offset)
        self.register_buffer('num_features', torch.tensor(len(categories)))

        self.emb_dim = torch.tensor(emb_dim)
        self.n_emb = self.categories.sum()
        self.linear = Linear(self.n_emb, self.emb_dim)
        self.register_buffer('scaler', torch.tensor(np.sqrt(self.n_emb)))
        self.add = MPAdd(0.5)
        
        self.bias = bias
        if self.bias:
            self.bias_linear = Linear(len(categories), self.emb_dim)
            self.register_buffer('bias_scaler', torch.tensor(np.sqrt(len(categories))))

    def forward(self, x):
  
        with torch.no_grad():
            x = x + self.categories_offset.unsqueeze(0)
            onehot = F.one_hot(x, num_classes=self.n_emb).to(torch.float32)
            onehot *= self.scaler 
        emb = self.linear(onehot) # emb has unit variance because we multiply by sqrt(dim_in)
        
        if self.bias:
            b = torch.arange(self.num_features, device=x.device)
            onehot = F.one_hot(b, num_classes=self.num_features).to(torch.float32)
            onehot *= self.bias_scaler
            bias = self.bias_linear(onehot)
            emb = self.add(emb, bias)
        
        return emb
    
    def get_all_feat_emb(self, feat_idx):
        emb_idx = torch.arange(self.categories[feat_idx], device=self.linear.weight.device) + self.categories_offset[feat_idx]
        onehot = F.one_hot(emb_idx, num_classes=self.n_emb).to(torch.float32)
        onehot *= self.scaler
        emb = self.linear(onehot)
        
        if self.bias:
            onehot = F.one_hot(torch.tensor(feat_idx, device=self.linear.weight.device), num_classes=self.num_features).to(torch.float32)
            onehot *= self.bias_scaler
            bias = self.bias_linear(onehot)
            emb = self.add(emb, bias)
        
        return emb



class MPSiLU(nn.Module):
    def forward(self, x):
        return F.silu(x) / 0.596    
    
    
class MPAdd(nn.Module):
    def __init__(self, t=0.5):
        super().__init__()
        self.t = t
        self.scaler = np.sqrt((1-t) ** 2 + t ** 2)
        
    def forward(self, a, b):
        return ((1-self.t) * a + self.t * b) / self.scaler
        
        
class MPCat(nn.Module):
    """
    From https://github.com/lucidrains/denoising-diffusion-pytorch/blob/main/denoising_diffusion_pytorch/karras_unet.py#L77
    """
    def __init__(self, t = 0.5, dim = -1):
        super().__init__()
        self.t = t
        self.dim = dim

    def forward(self, a, b):
        Na, Nb = a.shape[self.dim], b.shape[self.dim]
        C = np.sqrt((Na + Nb) / ((1. - self.t) ** 2 + self.t ** 2))

        a = a * (1. - self.t) / np.sqrt(Na)
        b = b * self.t / np.sqrt(Nb)

        return C * torch.cat((a, b), dim = self.dim)


class MPEmbedding(nn.Module):
    """
    Embedding corresponding to magnitude-preserving setup in Karras. Derives a one-hot vector from an index, scales it to unit variance, and then applies a magnitude-preserving linear layer.
    """
    def __init__(self, n_emb, emb_dim):
        super().__init__()
        self.linear = Linear(n_emb, emb_dim)
        self.n_emb = n_emb
        self.emb_dim = emb_dim
        self.register_buffer('scaler', torch.tensor(np.sqrt(n_emb)))

    def forward(self, x):
        
        with torch.no_grad():
            onehot = F.one_hot(x, num_classes=self.n_emb).to(torch.float32)
            onehot *= self.scaler
        out = self.linear(onehot)
        return out
    
    

class ConditionEmbedding(nn.Module):
    """Generates timestep embedding and combines with class embedding or other conditioning information to create conditioning embedding."""
    def __init__(self, time_emb_dim, out_dim, num_y_classes = None):
        super().__init__()
        
        if num_y_classes is not None: self.label_embedder = MPEmbedding(num_y_classes, out_dim)
        self.fourier_emb = MPFourierEmbedding(time_emb_dim)
        self.linear = Linear(time_emb_dim, out_dim)
        self.add = MPAdd(0.5)
        self.silu = MPSiLU()

        
    def forward(self, t, c = None):
        t_emb = self.linear(self.fourier_emb(t))
        if c is not None:
            t_emb = self.add(t_emb, self.label_embedder(c))
    
        return self.silu(t_emb)


class Gain(nn.Module):
    def __init__(self):
        super().__init__()
        self.gain = nn.Parameter(torch.tensor(0.))

    def forward(self, x):
        return x * self.gain
    

class FinalLayer(nn.Module):
    
    def __init__(self, dim_in, categories, num_cont_features, learn_gain=True, proportions=None):
        super().__init__()
        self.dim_in = dim_in
        self.num_cont_features = num_cont_features
        self.num_cat_features = len(categories)
        self.learn_gain = learn_gain

        # dim_out = self.num_cont_features + sum(categories)
        if self.learn_gain:
            self.gain = nn.Parameter(torch.zeros(1, num_cont_features + sum(categories)))
        if proportions is not None:
            cont_bias_init = torch.zeros((num_cont_features,))
            cat_bias_init = torch.cat(proportions).log()
            bias_init = torch.cat((cont_bias_init, cat_bias_init))
            self.init_bias = nn.Parameter(bias_init, requires_grad=False) 
        else:
            self.init_bias = None

        self.linear = Linear(dim_in, self.num_cont_features + sum(categories), concat_bias=True)
        self.split_chunks = [self.num_cont_features, *categories]
        self.cat_idx = 0
        if self.num_cont_features > 0:
            self.cat_idx = 1

            
    def forward(self, x):

        x = self.linear(x)
        
        if self.learn_gain:
            x = x * self.gain
 
        if self.init_bias is not None:
            x = x + self.init_bias

        out = torch.split(x, self.split_chunks, dim=-1)

        if self.num_cont_features > 0:
            cont_logits = out[0]
        else:
            cont_logits = None
        if self.num_cat_features > 0:
            cat_logits = out[self.cat_idx:]
        else:
            cat_logits = None
        
        return cat_logits, cont_logits

    
def exists(x):
    return x is not None

def default(val, d):
    if exists(val):
        return val
    return d() if callable(d) else d
    
    
 
class MPMLP(nn.Module):
    """
    Magnitude preserving MLP inspired by Karras 2023.
    """
    def __init__(self, emb_dim, cat_emb_dim, num_cont_features, categories, num_y_classes,
                 n_layers, n_units, proportions=None, feat_spec_cond=False):
        super().__init__()
        
        self.num_cont_features = num_cont_features
        self.num_cat_features = len(categories)
        self.num_features = self.num_cont_features + self.num_cat_features
        self.emb_dim = emb_dim

        self.cond_embedder = ConditionEmbedding(emb_dim, emb_dim, num_y_classes)

        in_dims = [emb_dim] + (n_layers-1) * [n_units]
        out_dims = n_layers * [n_units]
        layers = nn.ModuleList()
        for i in range(len(in_dims)):
            layers.append(Linear(in_dims[i], out_dims[i]))
            layers.append(MPSiLU()) 
        self.fc = nn.Sequential(*layers)
    
        self.final_layer = FinalLayer(out_dims[-1], categories, num_cont_features, proportions=proportions)
            
        # proj to embedding and additional add, cat layers (magnitude preserving)
        dim_in = (self.num_cat_features) * cat_emb_dim + num_cont_features
        self.proj = Linear(dim_in, emb_dim, concat_bias=False)
        self.add = MPAdd(0.5)
        self.cat = MPCat()
        
    @property
    def device(self):
        return next(self.parameters()).device
        
    def forward(self, x_cat_emb_t, x_cont_t, time, c = None):
            
        cond_emb = self.cond_embedder(time, c)
        x = self.cat(rearrange(x_cat_emb_t, 'B F D -> B (F D)'), x_cont_t)
        emb = self.add(self.proj(x), cond_emb) 
        h = self.fc(emb)
        cat_logits, cont_logits = self.final_layer(h) 

        return cat_logits, cont_logits
    

class MPWeightNetwork(nn.Module):
    def __init__(self, emb_dim):
        super().__init__()
        self.fourier = MPFourierEmbedding(emb_dim)
        self.fc = Linear(emb_dim, 1, concat_bias=True)
        self.gain = nn.Parameter(torch.tensor(0.0))
        
    def forward(self, u):
        c_noise = u.to(torch.float32).log() / 4
        x = self.fourier(c_noise)
        return torch.exp(self.fc(x).squeeze() * self.gain)
    
    def loss_fn(self, preds, avg_loss):
        # learn to fit expected average loss
        return (preds - avg_loss) ** 2
    
    

def compute_weight_magnitudes(weight_vector):
    return (1/weight_vector.shape[1] * (weight_vector ** 2).sum(1)).max().item()


def log_score_model_weights(writer, step, diff_model):
    
        model = diff_model.model

        layers = [model.cond_embedder.linear.weight,
                  model.proj.weight]

        mlp_weights = [model.fc[i].weight for i in range(len(model.fc)) if i % 2 == 0]
        mlp_labels = [f'mlp_{i}' for i in range(len(model.fc)) if i % 2 == 0]
        layers += mlp_weights
        layers.append(model.final_layer.linear.weight)
        
        labels = ['cond_embedder', 'proj'] 
        labels += mlp_labels
        labels.append('final_layer')

        for i, W in enumerate(layers):
            writer.add_scalar('layer/{}'.format(labels[i]), compute_weight_magnitudes(W), global_step = step)
            
