import torch
from torch import nn
import torch.nn.functional as F
from tqdm.autonotebook import tqdm
from train_utils import low_discrepancy_sampler
from timewarping import Timewarp, Timewarp_Logistic
from layers.adaptive_weighting import WeightNetwork
from layers.mlp import CatEmbedding
from layers.improved_cdtd import MPCatEmbedding, MPWeightNetwork
import matplotlib.pyplot as plt


class MixedTypeDiffusion(nn.Module):
    
    def __init__(self, model, dim, categories, num_features, task, 
                 sigma_data_cat, sigma_data_cont,
                 sigma_min_cat, sigma_max_cat, 
                 sigma_min_cont, sigma_max_cont, proportions,
                 cat_emb_init_sigma, 
                 timewarp_variant, timewarp_type,
                 timewarp_bins, timewarp_weight_low_noise=1.0,
                 timewarp_decay=0.0, normalize_losses=True,
                 adaptive_weight=False, cat_bias=False,
                 mp_cat=False,
                 simulate_missings=False):
        
        super(MixedTypeDiffusion, self).__init__()
        
        self.task = task
        self.dim = dim
        self.num_features = num_features
        self.num_cat_features = len(categories)
        self.num_cont_features = num_features - self.num_cat_features
        self.num_unique_cats = sum(categories)
        self.categories = categories
        self.normalize_losses = normalize_losses
        self.model = model
        self.simulate_missings = simulate_missings

        if mp_cat:
            self.cat_emb = MPCatEmbedding(dim, categories, bias=cat_bias)
        else:
            self.cat_emb = CatEmbedding(dim, categories, cat_emb_init_sigma, bias=cat_bias)

        self.register_buffer('sigma_data_cat', torch.tensor(sigma_data_cat))
        self.register_buffer('sigma_data_cont', torch.tensor(sigma_data_cont))
        
        if self.normalize_losses:
            assert proportions is not None
            max_ce_losses = torch.tensor([-torch.sum(p * p.log()) for p in proportions])
            self.register_buffer('max_losses',  torch.cat((max_ce_losses, torch.ones((self.num_cont_features,)))))

        # timewarping
        self.timewarp_type = timewarp_type
        self.sigma_min_cat = torch.tensor(sigma_min_cat)
        self.sigma_max_cat = torch.tensor(sigma_max_cat)
        self.sigma_min_cont = torch.tensor(sigma_min_cont)
        self.sigma_max_cont = torch.tensor(sigma_max_cont)

        # combine sigma boundaries for transforming sigmas to [0,1]
        sigma_min = torch.cat((
            torch.tensor(sigma_min_cat).repeat(self.num_cat_features),
            torch.tensor(sigma_min_cont).repeat(self.num_cont_features)
            ), dim = 0)
        sigma_max = torch.cat((
            torch.tensor(sigma_max_cat).repeat(self.num_cat_features),
            torch.tensor(sigma_max_cont).repeat(self.num_cont_features)
            ), dim = 0)
        self.register_buffer('sigma_max', sigma_max)
        self.register_buffer("sigma_min", sigma_min)
    
        self.timewarp_variant = timewarp_variant
        if self.timewarp_variant == 'cdcd':
            self.timewarp_cdf = Timewarp(self.timewarp_type, self.num_cat_features, self.num_cont_features, sigma_min, sigma_max, decay=timewarp_decay, num_bins=timewarp_bins)
        else:
            self.timewarp_cdf = Timewarp_Logistic(self.timewarp_type, self.num_cat_features, self.num_cont_features, sigma_min, sigma_max, weight_low_noise=timewarp_weight_low_noise, decay=timewarp_decay) 
            
        self.adaptive_weight = adaptive_weight
        if self.adaptive_weight:
            if mp_cat:
                self.weight_network = MPWeightNetwork(1024)
            else:
                self.weight_network = WeightNetwork(1024)

    @property
    def device(self):
        return next(self.model.parameters()).device
    
    
    def get_missing_mask(self, B, p=0.1):
        # 0 = missing, 1 = observed
        F = self.num_cont_features
        mask = torch.rand(B, F, device=self.device) > p 
        return mask.to(torch.float32)
        
          
    def diffusion_loss(self, x_cat_0, x_cont_0, cat_logits, cont_preds, sigma, cont_mask=None):
        
        assert len(cat_logits) == self.num_cat_features
        assert cont_preds.shape == x_cont_0.shape
        
        # cross entropy over categorical features for each individual
        ce_losses = torch.stack([F.cross_entropy(cat_logits[i], x_cat_0[:,i], reduction='none') for i in range(self.num_cat_features)], dim=1)
    
        # MSE loss over numerical features for each individual
        mse_losses = ((cont_preds - x_cont_0) ** 2)
        
        if cont_mask is not None:
            mse_losses *= cont_mask
        
        # compute EDM weight
        sigma_cont = sigma[:,self.num_cat_features:]
        cont_weight = (sigma_cont ** 2 + self.sigma_data_cont ** 2) / ((sigma_cont * self.sigma_data_cont) ** 2)

        weighted_losses = torch.cat((ce_losses, cont_weight * mse_losses), dim=1)
        unweighted_losses = torch.cat((ce_losses, mse_losses), dim=1) 
            
        if self.normalize_losses:
            weighted_losses /= self.max_losses
            unweighted_losses /= self.max_losses
              
        return unweighted_losses, weighted_losses
        
        
    def add_noise(self, x_cat_emb_0, x_cont_0, sigma):
        
        sigma_cat = sigma[:,:self.num_cat_features]
        sigma_cont = sigma[:,self.num_cat_features:]
    
        x_cat_emb_t = x_cat_emb_0 + torch.randn_like(x_cat_emb_0) * sigma_cat.unsqueeze(2)
        x_cont_t = x_cont_0 + torch.randn_like(x_cont_0) * sigma_cont
        
        return x_cat_emb_t, x_cont_t
        
 
    def loss_fn(self, x_cat, x_cont, y_cond, return_sigma=False, u=None):
           
        batch = x_cat.shape[0] if x_cat is not None else x_cont.shape[0]
        
        # get ground truth data
        x_cat_emb_0 = self.cat_emb(x_cat)        
        x_cont_0 = x_cont
        x_cat_0 = x_cat
        
        # draw u and convert to standard deviations for noise
        with torch.no_grad():
            if u is None:
                u = low_discrepancy_sampler(batch, device=self.device) # (B,)
            sigma = self.timewarp_cdf(u, invert = True).detach().to(torch.float32)
            u = u.to(torch.float32)
            assert sigma.shape == (batch, self.num_features)
            
        x_cat_emb_t, x_cont_t = self.add_noise(x_cat_emb_0, x_cont_0, sigma)
        
        if self.simulate_missings is not None:
            # set missing cont features to zero
            cont_mask = self.get_missing_mask(batch)
            x_cont_t *= cont_mask
        else:
            cont_mask = None
                       
        # retrieve score function evaluation
        cat_logits, cont_preds = self.precondition(x_cat_emb_t, 
                                                    x_cont_t,
                                                    y_cond,
                                                    u,
                                                    sigma,
                                                    )
        
        unweighted_losses, weighted_losses = self.diffusion_loss(x_cat_0, x_cont_0, cat_logits, 
                                                                 cont_preds, sigma, cont_mask=cont_mask)
        
        # adjust by adaptive weighting to achieve uniform weight of avg loss over timesteps
        if self.adaptive_weight:
            pred_avg_loss = self.weight_network(u) 
            weightnet_loss = self.weight_network.loss_fn(pred_avg_loss, weighted_losses.detach().mean(1))
            weighted_losses = weighted_losses / pred_avg_loss.detach().unsqueeze(1)
        else:
            weightnet_loss = None
   
        if return_sigma:
            return unweighted_losses, weighted_losses, weightnet_loss, sigma
        
        return unweighted_losses, weighted_losses

 
    def precondition(self, x_cat_emb_t, x_cont_t, y_cond, u, sigma):
        """
        Improved preconditioning proposed in the paper "Elucidating the Design 
        Space of Diffusion-Based Generative Models" (EDM) adjusted for categorical data
        """

        sigma_cat = sigma[:,:self.num_cat_features]
        sigma_cont = sigma[:,self.num_cat_features:]
        
        c_in_cat = 1 / (self.sigma_data_cat ** 2 + sigma_cat.unsqueeze(2) ** 2).sqrt() # batch, num_features, 1
        c_in_cont = 1 / (self.sigma_data_cont ** 2 + sigma_cont ** 2).sqrt()
        c_noise = u.log() / 4
            
        cat_logits, cont_preds = self.model(
                                     c_in_cat * x_cat_emb_t,
                                     c_in_cont * x_cont_t,
                                     c_noise, 
                                     c = y_cond,
                                     )
        
        assert len(cat_logits) == self.num_cat_features
        assert cont_preds.shape == x_cont_t.shape
        
        # apply preconditioning to continuous features
        c_skip = self.sigma_data_cont ** 2 / (sigma_cont ** 2 + self.sigma_data_cont ** 2)
        c_out = sigma_cont * self.sigma_data_cont / (sigma_cont ** 2 + self.sigma_data_cont ** 2).sqrt()
        D_x = c_skip * x_cont_t + c_out * cont_preds

        return cat_logits, D_x
        
        
    def score_interpolation(self, x_cat_emb_t, cat_logits, sigma, return_probs=False):
    
        if return_probs:
            # transform logits for categorical features to probabilities
            probs = []
            for logits in cat_logits:
                probs.append(F.softmax(logits.to(torch.float64), dim=1))
            return probs
           
        def interpolate_emb(i):
            p = F.softmax(cat_logits[i].to(torch.float64), dim=1)
            true_emb = self.cat_emb.get_all_feat_emb(i).to(torch.float64)
            return torch.matmul(p, true_emb)
        
        # take prob-weighted average of normalized ground truth embeddings
        x_cat_emb_0_hat = torch.zeros_like(x_cat_emb_t, device=self.device, dtype=torch.float64)
        for i in range(self.num_cat_features):
            x_cat_emb_0_hat[:,i,:] = interpolate_emb(i)

        # plug interpolated embedding into score function to interpolate score
        sigma_cat = sigma[:,:self.num_cat_features]
        interpolated_score = (x_cat_emb_t - x_cat_emb_0_hat) / sigma_cat.unsqueeze(2)
        
        return interpolated_score, x_cat_emb_0_hat
    
    
    @torch.inference_mode()
    def sampler_adj(self, cat_latents, cont_latents, y_dist=None, num_steps=200, verbose=False):
        
        B = cont_latents.shape[0] if self.num_cont_features > 0 else cat_latents.shape[0]
    
        # in case we condition on label y, sample y first from empirical distribution
        if y_dist is not None:
            y_gen = torch.multinomial(torch.tensor(y_dist).to(self.device), cont_latents.shape[0], replacement=True)
        else:
            y_gen = None
        
        # construct time steps
        u_steps = torch.linspace(1, 0, num_steps + 1, device=self.device, dtype=torch.float64)
        t_steps = self.timewarp_cdf(u_steps, invert=True)
        # assert (t_steps[0].to(torch.float32) == self.sigma_max).all()
        # assert (t_steps[:-1][-1].to(torch.float32) == self.sigma_min).all()
        
        # take last u step (=0) + 1e-7 to avoid taking log of 0 when computing c_noise
        # during training we clamp u to be in [0+1e-7, 1-1e-7] anyways
        # u_steps[-1] = 1e-7 
        # we do NOT need that, because the final u value passed to NN is u_steps[:-1][-1] not zero! 
        # the final step goes INTO t = 0, sigma = sigma_min, we never pass that to the model directly
        
        # initialize latents at maximum noise level
        t_cat_next = t_steps[0,:self.num_cat_features]
        t_cont_next = t_steps[0,self.num_cat_features:]
        x_cat_next = cat_latents.to(torch.float64) * t_cat_next.unsqueeze(1) 
        x_cont_next = cont_latents.to(torch.float64) * t_cont_next
        
        
        if verbose:
            print("=========== INIT ============")
            print(f"std = {x_cont_next.std(0).mean()}")
            
            
        for i, (t_cur, t_next, u_cur) in enumerate(zip(t_steps[:-1], t_steps[1:], u_steps[:-1])):
            
            t_cur = t_cur.repeat((B, 1))
            t_next = t_next.repeat((B, 1))
            t_cont_cur = t_cur[:, self.num_cat_features:]
            
            # get score model output
            cat_logits, x_cont_denoised = self.precondition(
                                                    x_cat_emb_t = x_cat_next.to(torch.float32),
                                                    x_cont_t = x_cont_next.to(torch.float32),
                                                    y_cond = y_gen,
                                                    u = u_cur.to(torch.float32).repeat((B,)),
                                                    sigma = t_cur.to(torch.float32),
                                                    )
            
            # estimate scores
            d_cat_cur, _ = self.score_interpolation(x_cat_next, cat_logits, t_cur)
            d_cont_cur = (x_cont_next - x_cont_denoised.to(torch.float64)) / t_cont_cur
            
            # adjust data samples
            h = t_next - t_cur
            x_cat_next = x_cat_next + h[:, :self.num_cat_features].unsqueeze(2) * d_cat_cur
            x_cont_next = x_cont_next + h[:, self.num_cat_features:] * d_cont_cur
            
            if verbose:
                print(f"=========== step {i} ============")
                print(f"std = {x_cont_next.std(0).mean()}")

        # final prediction of classes for categorical feature
        u_final = u_steps[:-1][-1]
        t_final = t_steps[:-1][-1].repeat(B,1)
        
        cat_logits, _ = self.precondition(x_cat_emb_t = x_cat_next.to(torch.float32),
                                          x_cont_t = x_cont_next.to(torch.float32),
                                          y_cond = y_gen, 
                                          u = u_final.to(torch.float32).repeat((B,)),
                                          sigma = t_final.to(torch.float32),
                                         )   
        
        # get probabilities for each category and derive generated classes
        probs = self.score_interpolation(x_cat_next, cat_logits, t_final, return_probs=True)
        x_cat_gen = torch.empty(B, self.num_cat_features, device=self.device)
        for i in range(self.num_cat_features):
            x_cat_gen[:,i] = probs[i].argmax(1)
            
        y_gen = y_gen.cpu() if y_gen is not None else None
        return x_cat_gen.cpu(), x_cont_next.cpu(), y_gen
    
    
    @torch.inference_mode()
    def sample_data(self, num_samples, batch_size, y_dist, num_steps, verbose=True):
        
        n_batches, remainder = divmod(num_samples, batch_size)
        sample_sizes = n_batches * [batch_size] + [remainder] if remainder != 0 else n_batches * [batch_size]

        x_cat_list = []
        x_cont_list = []
        y_list = []
        
        for num_samples in tqdm(sample_sizes, disable=(not verbose)):
            cat_latents = torch.randn((num_samples, self.num_cat_features, self.dim), device=self.device)
            cont_latents = torch.randn((num_samples, self.num_cont_features), device=self.device)
            x_cat_gen, x_cont_gen, y_gen = self.sampler_adj(cat_latents, cont_latents, y_dist, num_steps) # gives exact the same result as sampler but with fewer computations (define steps in sigma directly)
            x_cat_list.append(x_cat_gen)
            x_cont_list.append(x_cont_gen)      
            if y_gen is not None: y_list.append(y_gen) 
            
        x_cat = torch.cat(x_cat_list).cpu()
        x_cont = torch.cat(x_cont_list).cpu()
        y = torch.cat(y_list).cpu() if len(y_list) > 0 else None
          
        return x_cat, x_cont, y
        
        
    @torch.inference_mode()
    def impute_data(self, X_cat, X_cont, y, mask_obs, batch_size, num_steps, resample_steps=10, verbose=True):
        
        B = X_cat.shape[0] if X_cat is not None else X_cont.shape[0]
        data_split = torch.split(torch.arange(B), batch_size)
        sample_sizes = [d.shape[0] for d in data_split]
        
        x_cat_list = []
        x_cont_list = []
        y_list = []
        
        for i, num_samples in tqdm(enumerate(sample_sizes), disable=(not verbose)):
            idx = data_split[i]
            mask_sample = mask_obs[idx]
            x_cat_sample = X_cat[idx].to(self.device)
            x_cont_sample = X_cont[idx].to(self.device)
            y_sample = y[idx].to(self.device) if y is not None else None
            cat_latents = torch.randn((num_samples, self.num_cat_features, self.dim), device=self.device)
            cont_latents = torch.randn((num_samples, self.num_cont_features), device=self.device)
            x_cat_gen, x_cont_gen, y_gen = self.imputer(x_cat_sample, x_cont_sample, y_sample, 
                                                        mask_sample, cat_latents, 
                                                        cont_latents, num_steps, verbose=False,
                                                        resample_steps=resample_steps)
            x_cat_list.append(x_cat_gen.cpu())
            x_cont_list.append(x_cont_gen.cpu())      
            if y_gen is not None: y_list.append(y_gen.cpu()) 
            
        x_cat = torch.cat(x_cat_list)
        x_cont = torch.cat(x_cont_list)
        y = torch.cat(y_list) if len(y_list) > 0 else None
          
        return x_cat, x_cont, y
    
    
    @torch.inference_mode()
    def imputer(self, x_cat, x_cont, y, mask, cat_latents, cont_latents, num_steps=200,
                resample_steps=20, verbose=False, ode=True):
        
        B = cont_latents.shape[0] if self.num_cont_features > 0 else cat_latents.shape[0]

        # construct time steps
        u_steps = torch.linspace(1, 0, num_steps + 1, device=self.device, dtype=torch.float64)
        t_steps = self.timewarp_cdf(u_steps, invert=True)
        
        # construct observed xs
        x_cat_emb_true = self.cat_emb(x_cat).to(torch.float64)
        x_cont_true = x_cont.to(torch.float64)
        mask_cat = mask[:, :self.num_cat_features]
        mask_cont = mask[:, self.num_cat_features:]
        
        # initialize latents at maximum noise level 
        t_cat_next = t_steps[0,:self.num_cat_features]
        t_cont_next = t_steps[0,self.num_cat_features:]
        x_cat_next = cat_latents.to(torch.float64) * t_cat_next.unsqueeze(1) 
        x_cont_next = cont_latents.to(torch.float64) * t_cont_next
        
        # condition on observed data
        x_cat_emb_true_noisy = x_cat_emb_true + t_cat_next.unsqueeze(1) * torch.randn_like(x_cat_emb_true)
        x_cat_next = mask_cat.unsqueeze(2) * x_cat_emb_true_noisy + (1-mask_cat).unsqueeze(2) * x_cat_next
        x_cont_true_noisy = x_cont_true  + t_cont_next * torch.randn_like(x_cont_true) 
        x_cont_next = mask_cont * x_cont_true_noisy + (1-mask_cont) * x_cont_next
    
        for i, (t_cur, t_next, u_cur) in enumerate(zip(t_steps[:-1], t_steps[1:], u_steps[:-1])):
            
            t_cur = t_cur.repeat((B, 1))
            t_next = t_next.repeat((B, 1))
            h = t_next - t_cur
            t_cat_cur = t_cur[:,:self.num_cat_features]
            t_cont_cur = t_cur[:, self.num_cat_features:]
            t_cat_next = t_next[:, :self.num_cat_features]
            t_cont_next = t_next[:, self.num_cat_features:]
            forward_step_std = (t_cur ** 2 - t_next ** 2).sqrt()
            forward_step_std_cat = forward_step_std[:, :self.num_cat_features]
            forward_step_std_cont = forward_step_std[:, self.num_cat_features:]
            
            for u in range(resample_steps+1):
                # resampling steps to improve harmonization of the data
            
                # get score model output
                cat_logits, x_cont_denoised = self.precondition(
                                                        x_cat_emb_t = x_cat_next.to(torch.float32),
                                                        x_cont_t = x_cont_next.to(torch.float32),
                                                        y_cond = y,
                                                        u = u_cur.to(torch.float32).repeat((B,)),
                                                        sigma = t_cur.to(torch.float32),
                                                        )
                
                # estimate scores
                d_cat_cur, _ = self.score_interpolation(x_cat_next, cat_logits, t_cur)
                d_cont_cur = (x_cont_next - x_cont_denoised.to(torch.float64)) / t_cont_cur
                
                # adjust data samples
                if ode:
                    x_cat_next = x_cat_next + h[:, :self.num_cat_features].unsqueeze(2) * d_cat_cur
                    x_cont_next = x_cont_next + h[:, self.num_cat_features:] * d_cont_cur
                else:
                    dt = u_steps[i+1] - u_cur
                    dsigma = t_next - t_cur
                    dsigmadt = dsigma / dt
                    x_cat_next = x_cat_next + (2 * h[:, :self.num_cat_features].unsqueeze(2) * d_cat_cur) + torch.sqrt(2 * dsigmadt[:, :self.num_cat_features] * t_cat_cur).unsqueeze(2) * torch.randn_like(x_cat_next)
                    x_cont_next = x_cont_next + (2 * h[:, self.num_cat_features:] * d_cont_cur) + torch.sqrt(2 * dsigmadt[:, self.num_cat_features:] * t_cont_cur) * torch.randn_like(x_cont_next)

                # condition on true data at that timestep (t-1)
                x_cat_emb_true_noisy = x_cat_emb_true + t_cat_next.unsqueeze(2) * torch.randn_like(x_cat_emb_true)
                x_cont_true_noisy = x_cont_true + t_cont_next * torch.randn_like(x_cont_true) 
                x_cat_next = mask_cat.unsqueeze(2) * x_cat_emb_true_noisy + (1-mask_cat).unsqueeze(2) * x_cat_next
                x_cont_next = mask_cont * x_cont_true_noisy + (1-mask_cont) * x_cont_next
                
                # resampling (t-1 --> t) to harmonize the data
                if u < resample_steps and u_cur > u_steps[:-1][-1]:
                    x_cat_next = x_cat_next + forward_step_std_cat.unsqueeze(2) * torch.randn_like(x_cat_next)
                    x_cont_next = x_cont_next + forward_step_std_cont * torch.randn_like(x_cont_next)
                    
            if verbose:
                print(f"=========== step {i} ============")
                print(f"std = {x_cont_next.std(0).mean()}")
                theo_std = (1 + t_next[0,self.num_cat_features:] ** 2).sqrt()
                print(f"theoretical std = {theo_std}")
            

        # final prediction of classes for categorical feature
        u_final = u_steps[:-1][-1]
        t_final = t_steps[:-1][-1].repeat(B,1)
        
        cat_logits, _ = self.precondition(x_cat_emb_t = x_cat_next.to(torch.float32),
                                          x_cont_t = x_cont_next.to(torch.float32),
                                          y_cond = y, 
                                          u = u_final.to(torch.float32).repeat((B,)),
                                          sigma = t_final.to(torch.float32),
                                         )   
        
        # get probabilities for each category and derive generated classes
        probs = self.score_interpolation(x_cat_next, cat_logits, t_final, return_probs=True)
        x_cat_gen = torch.empty(B, self.num_cat_features, device=self.device)
        for i in range(self.num_cat_features):
            x_cat_gen[:,i] = probs[i].argmax(1)
            
        # combine imputations with true data
        x_cat_gen = mask_cat * x_cat + (1-mask_cat) * x_cat_gen
        x_cont_gen = mask_cont * x_cont_true + (1-mask_cont) * x_cont_next
            
        y_gen = y.cpu() if y is not None else None
        return x_cat_gen.cpu(), x_cont_gen.cpu(), y_gen
    
