import matplotlib.pyplot as plt
import torch
import imageio
import numpy as np
import matplotlib.colors as mcolors
import os
import pandas as pd
import seaborn as sns
from matplotlib.pyplot import cm
from einops import repeat
import matplotlib.patches as mpatches
import matplotlib.lines as mlines
      

def scale_t(t, t_min, t_max):
    return (t - t_min) / (t_max - t_min)


def plot_timewarp_cdf(log_path, diff_model, scale_x, log_scale=False):
    
    figname = "timewarp_cdf"
    
    with torch.no_grad():
        u = torch.linspace(0, 1, 500, device=diff_model.device)
        s = diff_model.timewarp_cdf(u, invert = True)
        y = diff_model.timewarp_cdf(s).cpu()
        
    if scale_x:
        x = ((s - diff_model.sigma_min) / (diff_model.sigma_max - diff_model.sigma_min)).cpu()
        xlabel = "Standard Deviation scaled to [0,1]"
        figname += "_scaled"
    else:
        x = s.cpu()
        xlabel = "Standard Deviation"
        
    if log_scale:
        xlabel += " (log scale)"
        figname += "_log"
        
    if diff_model.timewarp_type == 'single':
        x = x[:,0].unsqueeze(1)
    elif diff_model.timewarp_type == 'bytype':
        x = torch.stack((x[:,0], x[:,-1]), dim=1)        
    elif diff_model.timewarp_type == 'single_cont':
        x = x[:, :diff_model.num_cat_features + 1]
        
    labels = [f"{i}" for i in range(s.shape[1])]
    colors = cm.rainbow(np.linspace(0, 1, s.shape[1]))
    
    fig = plt.figure()
    for i in range(x.shape[1]):
        plt.plot(x[:,i], y[:,i], marker='none', linestyle='-', linewidth=1, color=colors[i], label=labels[i])
    if log_scale:
        plt.xscale('log')
    plt.legend(bbox_to_anchor=(0.5, 1.1), loc='center', ncols=5, title='Feature')
    plt.xlabel(xlabel)
    plt.ylabel('Loss')
    plt.tight_layout()
    
    if log_path is None:
        return fig
    else:
        fig.savefig(os.path.join(log_path, figname + '.png'), dpi=300, bbox_inches='tight')
        plt.close()
        

def plot_timewarp_pdf(log_path, diff_model, scale_x, log_scale=False):
    
    figname = "timewarp_pdf"

    with torch.no_grad():
        u = torch.linspace(1e-7, 1-1e-7, 500, device=diff_model.device)
        s = diff_model.timewarp_cdf(u, invert = True)
        y = diff_model.timewarp_cdf(s, return_pdf = True, normalize = True).cpu()
    
    if scale_x:
        xlabel = "Standard Deviation scaled to [0,1]"
        x = ((s - diff_model.sigma_min) / (diff_model.sigma_max - diff_model.sigma_min)).cpu()
        figname += "_scaled"
    else:
        xlabel = "Standard Deviation"
        x = s.cpu()
        
    # add point at (0,0)
    x = torch.cat((torch.zeros((1, diff_model.num_features)), x), dim=0)
    y = torch.cat((torch.zeros((1, y.shape[1])), y), dim=0)
        
    if log_scale:
        xlabel += " (log scale)"
        figname += "_log"
        
    if diff_model.timewarp_type == 'single':
        x = x[:,0].unsqueeze(1)
    elif diff_model.timewarp_type == 'bytype':
        x = torch.stack((x[:,0], x[:,-1]), dim=1)        
    elif diff_model.timewarp_type == 'single_cont':
        x = x[:, :diff_model.num_cat_features + 1]
            
    labels = [f"{i}" for i in range(s.shape[1])]
    # colors = [mcolors.to_hex(plt.cm.tab20(i)) for i in range(diff_model.num_features)]
    # colors = plt.rcParams['axes.prop_cycle'].by_key()['color']
    colors = cm.rainbow(np.linspace(0, 1, s.shape[1]))
    
    fig = plt.figure()
    for i in range(x.shape[1]):
        plt.plot(x[:,i], y[:,i], marker='none', linestyle='-', linewidth=1, color=colors[i], label=labels[i])
    # plt.legend(bbox_to_anchor=(0.5, 1.1), loc='center', ncols=5, title='Feature')
    if log_scale:
        plt.xscale('log')
    plt.xlabel(xlabel)
    plt.tight_layout()
    
    # plt.xscale('log')
    if log_path is None:
        return fig
    else:
        fig.savefig(os.path.join(log_path, figname + '.png'), dpi=300, bbox_inches='tight')
        plt.close()
        
    
def plot_timewarp_inverse_cdf(log_path, diff_model, scale_y):
    
    with torch.no_grad():
        u = torch.linspace(0, 1, 800, device=diff_model.device)
        s = diff_model.timewarp_cdf(u, invert = True)
    u = u.cpu()
    if scale_y:
        y = ((s - diff_model.sigma_min) / (diff_model.sigma_max - diff_model.sigma_min)).cpu()
        ylabel = "Standard Deviation scaled to [0,1]"
    else:
        y = s.cpu()
        ylabel = "Standard Deviation"

    
    if diff_model.timewarp_type == 'single':
        y = y[:,0].unsqueeze(1)
    elif diff_model.timewarp_type == 'bytype':
        y = torch.stack((y[:,0], y[:,-1]), dim=1)        
    elif diff_model.timewarp_type == 'single_cont':
        y = y[:, :diff_model.num_cat_features + 1]
        
    labels = [f"{i}" for i in range(s.shape[1])]
    colors = cm.rainbow(np.linspace(0, 1, s.shape[1]))
    
    fig = plt.figure()
    for i in range(y.shape[1]):
        plt.plot(u, y[:,i], marker='none', linestyle='-', linewidth=1, color=colors[i], label=labels[i])
    plt.legend(bbox_to_anchor=(0.5, 1.1), loc='center', ncols=5, title='Feature')
    plt.xlabel('Global Time')
    plt.ylabel(ylabel)
    
    if log_path is None:
        return fig
    else:
        fig.savefig(os.path.join(log_path, 'timewarp_cdf_inverse.png'), dpi=300, bbox_inches='tight')
        plt.close()
    

def plot_loss_gif_frame(diff_model):
    
    fig = plot_timewarp_cdf(None, diff_model, scale_x=True)
    fig.canvas.draw()
    mat = np.array(fig.canvas.renderer._renderer)
    mat = imageio.core.util.Array(mat)
    plt.close()
    
    return mat


def plot_diff_in_means(log_path, diff_in_means, num_rows=16, iteration=None):
    
    # bring data into image shape, num_rows * num_rows
    size = num_rows ** 2
    if len(diff_in_means) < size:
        diff_in_means = np.concatenate((diff_in_means, np.array([np.nan] * (size - len(diff_in_means)))))
    x = diff_in_means.reshape((num_rows, num_rows))
    
    fig = plt.figure()
    plt.matshow(x, fig, cmap = 'viridis', vmax = 1.0)#, vmin = 0, vmax = 2)
    plt.tick_params(axis='x', which='both', bottom=False, top=False, labeltop=False) 
    plt.tick_params(axis='y', which='both', left=False, right=False, labelleft=False) 
    cb = plt.colorbar()
    if iteration:
        plt.title(f'Iteration: {iteration}')
    
    if log_path is None:
        return fig
    else:
        fig.savefig(os.path.join(log_path, 'diff_in_means.png'), dpi=300, bbox_inches='tight')
        plt.close()
    

def plot_diff_in_means_gif_frame(diff_in_means, iteration=None):
    
    fig = plot_diff_in_means(None, diff_in_means, iteration=iteration)
    fig.canvas.draw()
    mat = np.array(fig.canvas.renderer._renderer)
    mat = imageio.core.util.Array(mat)
    plt.close()
    
    return mat



def plot_diff_in_corr(log_path, diff_in_corr, prefix=None, iteration=None, vmax=0.1):
    
    fig = plt.figure()
    plt.matshow(diff_in_corr, fig, cmap='Reds', vmin=0, vmax=vmax)
    plt.tick_params(axis='x', which='both', bottom=False, top=False, labeltop=False) 
    plt.tick_params(axis='y', which='both', left=False, right=False, labelleft=False) 
    cb = plt.colorbar()
    if iteration:
        plt.title(f'Iteration: {iteration}')

    if log_path is None:
        return fig
    else:
        label = 'diff_in_corr.png'
        if prefix is not None:
            label = prefix + label
        fig.savefig(os.path.join(log_path, label), dpi=300, bbox_inches='tight')
        plt.close()
        
        
def plot_diff_in_corr_gif_frame(diff_in_corr, iteration=None):

    fig = plot_diff_in_corr(None, diff_in_corr, iteration=iteration)
    fig.canvas.draw()
    mat = np.array(fig.canvas.renderer._renderer)
    mat = imageio.core.util.Array(mat)
    plt.close()

    return mat


def generate_timewarp_gif(log_path, cdf_frames, uniform = False, fps=5):
    for _ in range(6):
        cdf_frames.append(cdf_frames[-1])
    
    if uniform:
        label = 'timewarp_cdf_uniform.gif'
    else:
        label = 'timewarp_cdf.gif'

    imageio.mimsave(os.path.join(log_path, label), cdf_frames, duration = 1000/fps)


def generate_diff_in_means_gif(log_path, frames, fps=5):
    for _ in range(6):
        frames.append(frames[-1])
    imageio.mimsave(os.path.join(log_path, 'diff_in_means.gif'), frames, duration = 1000/fps)


def generate_diff_in_corr_gif(log_path, frames, fps=5):
    for _ in range(6):
        frames.append(frames[-1])
    imageio.mimsave(os.path.join(log_path, 'diff_in_corr.gif'), frames, duration = 1000/fps)



def plot_train_loss(log_path, loss_hist, average_over_features = False, average_by_type = False, num_cat_features = None, eval_step = 50):
    
    labels = [f"feature {i}" for i in range(loss_hist.shape[1])]
    # colors = [mcolors.to_hex(plt.cm.tab20(i)) for i in range(loss_hist.shape[1])]
    colors = cm.rainbow(np.linspace(0, 1, loss_hist.shape[1]))
    plot_label = 'losses'
    if average_over_features:
        labels = ['total average loss']
        plot_label = 'avg_loss'
        loss_hist = loss_hist.mean(1).unsqueeze(1)
    if average_by_type:
        assert num_cat_features is not None
        labels = ['categorical', 'continuous']
        plot_label = 'avg_losses_by_type'
        loss_hist_cat = loss_hist[:,:num_cat_features].mean(1)
        loss_hist_cont = loss_hist[:,num_cat_features:].mean(1)
        loss_hist = torch.stack((loss_hist_cat, loss_hist_cont), 1)
        
    fig = plt.figure(figsize=(10,6))
    for i, label in enumerate(labels):
        plt.plot(loss_hist[:,i], label=label, color=colors[i])
    plt.xlabel(f'{eval_step}s of steps')
    plt.ylabel(f'Average Cross Entropy or MSE over last {eval_step} steps')
    plt.legend(bbox_to_anchor=(0.5, 1.1), loc='center', ncols=5)
    fig.savefig(os.path.join(log_path, f'{plot_label}.png'), dpi=300, bbox_inches='tight')
    plt.close()

def get_max_ce(X_cat, y, categories):
    
    if y is not None:
        # get CE for p(x_cat, x_cont | y) model
        y_vals, y_counts = y.unique(return_counts=True)
        y_props = y_counts / y_counts.sum()
        
        ce_vals = []
        for i, y_val in enumerate(y_vals):
            ces = []
            for j in range(len(categories)):
                _, counts = X_cat[y == y_val,j].unique(return_counts=True)
                proportions = counts / y_counts[i]
                ce = -sum(proportions * proportions.log())
                ces.append(ce)
            ce_vals.append(ces)
        # take expected value over y values
        worst_ces = (torch.tensor(ce_vals) * y_props.unsqueeze(1)).sum(dim=0).tolist()
            
    else:
        # get CE for p(x_cat, x_cont, y) model
        worst_ces = []
        n_sample = X_cat.shape[0]
        for i in range(len(categories)):
            uniq_vals, counts = X_cat[:,i].unique(return_counts=True)
            # assert len(uniq_vals) == categories[i] # causes problems if e.g. X_cat in validation set does not include all categories
            proportions = counts / n_sample
            # random classifier predicts majority classes
            ce = -sum(proportions * proportions.log())
            worst_ces.append(ce.item())

    return worst_ces
    
    

def plot_cdf_and_loss(save_path, diff_model, data_dict, num_cat_features, ax = None, prefix = ''):
    with torch.no_grad():
        losses = data_dict['unweighted_avg']
        s = data_dict['sigma'].to(diff_model.device)
        y = diff_model.timewarp_cdf(s).cpu()

        x = ((s - diff_model.sigma_min) / (diff_model.sigma_max - diff_model.sigma_min)).cpu()
        xlabel = "$\sigma_t$ scaled to [0,1]"
        
    if diff_model.timewarp_type == 'single':
        x = x[:,0].unsqueeze(1)
        loss = losses.mean(1).unsqueeze(1)
    elif diff_model.timewarp_type == 'bytype':
        x = torch.stack((x[:,0], x[:,-1]), dim=1)   
        mse_loss = losses[:, num_cat_features:].mean(1)
        ce_loss = losses[:, :num_cat_features].mean(1)
        loss = torch.stack((ce_loss, mse_loss), dim=1)
    else:
        loss = losses
        
    if x.shape[1] == 2:
        labels = ["categorical", 'continuous']
    else:
        labels = [f"{i}" for i in range(s.shape[1])]
        
    colors = cm.tab20.colors * 4
    
    if ax is None:
        fig, ax = plt.subplots()
        
    leg_colors = []   
    for i in range(x.shape[1]):
        ax.plot(x[:,i], y[:,i], marker='none', linestyle='-', linewidth=1, color=colors[i], label=labels[i])
        ax.plot(x[:,i], loss[:,i], marker='none', linestyle='--', linewidth=1, color=colors[i], label=labels[i])
        leg_colors.append(mpatches.Patch(color = colors[i], label = labels[i]))
        
    leg_lines = [mlines.Line2D([], [], linestyle='-', label='fitted function', color='grey'),
                mlines.Line2D([], [], linestyle='--', label='true diffusion loss', color='grey')]

    plt.legend(handles = leg_colors + leg_lines)
    plt.xlabel(xlabel)
    plt.ylabel('Loss')
    plt.xscale('log')
    plt.xlim(1e-3, 1)
    if save_path is not None:
        fig.savefig(os.path.join(save_path, prefix + 'cdf_fit.png'), dpi=300, bbox_inches='tight')
    else:
        return ax

    

def plot_weighting(save_path, diff_model, ax=None, prefix=''):
    with torch.no_grad():
        u = torch.linspace(1e-7, 1-1e-7, 500, device=diff_model.device)
        s = diff_model.timewarp_cdf(u, invert = True)
        y = diff_model.timewarp_cdf(s, return_pdf = True, normalize = True).cpu()
        x = ((s - diff_model.sigma_min) / (diff_model.sigma_max - diff_model.sigma_min)).cpu()

        # add point at (0,0)
        x = torch.cat((torch.zeros((1, diff_model.num_features)), x), dim=0) 
        y = torch.cat((torch.zeros((1, y.shape[1])), y), dim=0)

        if diff_model.timewarp_type == 'single':
            x = x[:,0].unsqueeze(1)
            labels = [f"{i}" for i in range(s.shape[1])]
        elif diff_model.timewarp_type == 'bytype':
            x = torch.stack((x[:,0], x[:,-1]), dim=1)
            labels = ['categorical', 'continuous']     
        else:
            labels = [f"{i}" for i in range(s.shape[1])]
        colors = cm.tab20.colors * 4

        if ax is None:
            fig, ax = plt.subplots()
        for i in range(x.shape[1]):
            ax.plot(x[:,i], y[:,i], marker='none', linestyle='-', linewidth=1, color=colors[i], label=labels[i])
        if y.shape[1] == 2:
            ax.legend()
        ax.set_xlabel("$\sigma_t$ scaled to [0,1]")
        ax.set_ylabel("$p(\sigma_t)$")
        
        if save_path is not None:
            fig.savefig(os.path.join(save_path, prefix + 'pdf_weighting.png'), dpi=300, bbox_inches='tight')
        else:
            return ax

    

def plot_cat_losses_wrt_sigma(save_path, categories, sigmas_cat, cat_losses, max_ce, prefix, scale_x):
    fig = plt.figure()
    label = 'ce_loss_wrt_sigma'
    # colors = list(mcolors.TABLEAU_COLORS)
    colors = cm.rainbow(np.linspace(0, 1, sigmas_cat.shape[1]))
    for i in range(sigmas_cat.shape[1]):
        plt.plot(sigmas_cat[:,i], cat_losses[:,i], marker='none', 
                 linestyle='-', linewidth=1, 
                 label=f"{i}: {categories[i]} cats", color=colors[i])
        plt.axhline(y=max_ce[i], linestyle='--', color=colors[i], linewidth=1)
    plt.xlabel('Standard Deviation of Noise')
    plt.ylabel('Avg. CEs (per obs.) over whole training set')
    plt.legend(bbox_to_anchor=(0.5, 1.1), loc='center', ncols=5, title='Feature')
    plt.tight_layout()
    if scale_x:
        plt.xscale('log')
        label += '_log_scale'
    fig.savefig(os.path.join(save_path, prefix + label + '.png'), dpi=300, bbox_inches='tight')
    plt.close()
    
    
def plot_cont_losses_wrt_sigma(save_path, sigmas_cont, mse_loss, prefix, scale_x):
    label = 'mse_loss_wrt_sigma'
    colors = cm.rainbow(np.linspace(0, 1, sigmas_cont.shape[1]))
    fig = plt.figure()
    for i in range(sigmas_cont.shape[1]):
        avg = mse_loss[:,i]
        # std = mse_loss_std[:,i]
        plt.plot(sigmas_cont[:,i], mse_loss[:,i], marker='none',
                 label=f"{i}", linestyle='-', linewidth=1, color=colors[i])
        # plt.fill_between(s_cont[:,i], avg-std, avg+std, alpha=0.2)
    plt.xlabel('Standard Deviation of Noise')
    plt.ylabel('Avg. MSEs (per obs.) over whole training set')
    plt.legend(bbox_to_anchor=(0.5, 1.1), loc='center', ncols=5, title='Feature')
    if scale_x:
        plt.xscale('log')
        label += '_log_scale'
    fig.savefig(os.path.join(save_path, prefix + label + '.png'), dpi=300, bbox_inches='tight')
    plt.close()


def plot_cat_losses_wrt_u(save_path, categories, us, cat_losses, prefix):
    colors = cm.rainbow(np.linspace(0, 1, cat_losses.shape[1]))
    fig = plt.figure()
    for i in range(len(categories)):
        plt.plot(us, cat_losses[:,i], marker='none', linestyle='-', linewidth=1, 
                 label=f"{categories[i]}", color=colors[i])
    plt.xlabel('Global time')
    plt.ylabel('Avg. CEs (per obs.) over whole training set')
    plt.tight_layout()
    plt.legend(bbox_to_anchor=(0.5, 1.1), loc='center', ncols=5, title='Categories')
    fig.savefig(os.path.join(save_path, prefix + 'ce_loss_wrt_u.png'), dpi=300, bbox_inches='tight')
    plt.close()


def plot_cont_losses_wrt_u(save_path, us, mse_loss, prefix):
    colors = cm.rainbow(np.linspace(0, 1, mse_loss.shape[1]))
    fig = plt.figure()
    for i in range(mse_loss.shape[1]):
        plt.plot(us, mse_loss[:,i], marker='none', label=f"{i}",
                 linestyle='-', linewidth=1, color=colors[i])
    plt.xlabel('Global time')
    plt.ylabel('Avg. MSEs (per obs.) over whole training set')
    plt.legend(bbox_to_anchor=(0.5, 1.1), loc='center', ncols=5, title='Feature')
    fig.savefig(os.path.join(save_path, prefix + 'mse_loss_wrt_u.png'), dpi=300, bbox_inches='tight')
    plt.close()


def plot_avg_losses_wrt_u(save_path, prefix, data_dict, num_cat_features):
    
    fig = plt.figure()
    plt.plot(data_dict['u'], data_dict['weighted_avg'].mean(1), marker='none', linestyle='-', linewidth=1)
    plt.fill_between(data_dict['u'], data_dict['weighted_avg'].mean(1)-data_dict['weighted_avg'].std(1),
                     data_dict['weighted_avg'].mean(1)+data_dict['weighted_avg'].std(1), alpha=0.1)
    plt.xlabel('Global time')
    plt.ylabel('Avg. total weighted loss (per feature and obs.) over train set')
    fig.savefig(os.path.join(save_path, prefix + 'avg_total_weighted_loss_wrt_u.png'), dpi=300)
    plt.close()
    
    fig = plt.figure()
    plt.plot(data_dict['u'], data_dict['unweighted_avg'].mean(1), marker='none', linestyle='-', linewidth=1)
    plt.fill_between(data_dict['u'], data_dict['unweighted_avg'].mean(1)-data_dict['unweighted_avg'].std(1),
                     data_dict['unweighted_avg'].mean(1)+data_dict['unweighted_avg'].std(1), alpha=0.1)
    plt.xlabel('Global time')
    plt.ylabel('Avg. total unweighted loss (per feature and obs.) over train set')
    fig.savefig(os.path.join(save_path, prefix + 'avg_total_unweighted_loss_wrt_u.png'), dpi=300)
    plt.close()

    # weighted losses
    
    ce_loss = data_dict['weighted_avg'][:, :num_cat_features]
    fig = plt.figure()
    plt.plot(data_dict['u'], ce_loss.mean(1),  marker='none', linestyle='-', linewidth=1)
    plt.fill_between(data_dict['u'], ce_loss.mean(1)-ce_loss.std(1), ce_loss.mean(1)+ce_loss.std(1), alpha=0.1)
    plt.xlabel('Global time')
    plt.ylabel('Avg. weighted CE loss (per feature and obs.) over train set')
    # plt.ylim(-0.5,4)
    fig.savefig(os.path.join(save_path, prefix + 'avg_weighted_ce_loss_wrt_u.png'), dpi=300)
    plt.close()

    mse_loss = data_dict['weighted_avg'][:, num_cat_features:]
    fig = plt.figure()
    plt.plot(data_dict['u'], mse_loss.mean(1),  marker='none', linestyle='-', linewidth=1)
    plt.fill_between(data_dict['u'], mse_loss.mean(1)-mse_loss.std(1), mse_loss.mean(1)+mse_loss.std(1), alpha=0.1)
    plt.xlabel('Global time')
    plt.ylabel('Avg. weighted MSE loss (per feature and obs.) over train set')
    fig.savefig(os.path.join(save_path, prefix + 'avg_weighted_mse_loss_wrt_u.png'), dpi=300)
    plt.close()
    
    # unweighted losses
    
    ce_loss = data_dict['unweighted_avg'][:, :num_cat_features]
    fig = plt.figure()
    plt.plot(data_dict['u'], ce_loss.mean(1),  marker='none', linestyle='-', linewidth=1)
    plt.fill_between(data_dict['u'], ce_loss.mean(1)-ce_loss.std(1), ce_loss.mean(1)+ce_loss.std(1), alpha=0.1)
    plt.xlabel('Global time')
    plt.ylabel('Avg. unweighted CE loss (per feature and obs.) over train set')
    # plt.ylim(-0.5,4)
    fig.savefig(os.path.join(save_path, prefix + 'avg_unweighted_ce_loss_wrt_u.png'), dpi=300)
    plt.close()

    mse_loss = data_dict['unweighted_avg'][:, num_cat_features:]
    fig = plt.figure()
    plt.plot(data_dict['u'], mse_loss.mean(1),  marker='none', linestyle='-', linewidth=1)
    plt.fill_between(data_dict['u'], mse_loss.mean(1)-mse_loss.std(1), mse_loss.mean(1)+mse_loss.std(1), alpha=0.1)
    plt.xlabel('Global time')
    plt.ylabel('Avg. unweighted MSE loss (per feature and obs.) over train set')
    fig.savefig(os.path.join(save_path, prefix + 'avg_unweighted_mse_loss_wrt_u.png'), dpi=300)
    plt.close()
    
    


def plot_cdf_losses_wrt_sigma(logdir, timewarp_type, sigmas, cdf_loss, num_cat_features, prefix):
    
    if num_cat_features > 0:
        if timewarp_type in ['single', 'bytype']:
            x = sigmas[:,0].unsqueeze(1)
            y = cdf_loss[:,0].unsqueeze(1)
        else:
            x = sigmas[:,:num_cat_features]
            y = cdf_loss[:,:num_cat_features]
            
        colors = cm.rainbow(np.linspace(0, 1, y.shape[1]))
        fig = plt.figure()
        for i in range(y.shape[1]):
            plt.plot(x[:,i], y[:,i], marker='none', linestyle='-', linewidth=1, color=colors[i])
        plt.xlabel('Standard Deviation of Noise')
        plt.ylabel('Avg. MSEs (per obs.) over whole train set (cat. features)')
        fig.savefig(os.path.join(logdir, prefix + 'timewarp_cat_loss.png'), dpi=300, bbox_inches='tight')
        plt.close()
            
    if sigmas.shape[1] - num_cat_features > 0:
        if timewarp_type == 'single':
            x = sigmas[:,0].unsqueeze(1)
            y = cdf_loss[:,0].unsqueeze(1)
        elif timewarp_type == 'bytype':
            x = sigmas[:,-1].unsqueeze(1)
            y = cdf_loss[:,-1].unsqueeze(1)
        else:
            x = sigmas[:,num_cat_features:]
            y = cdf_loss[:,num_cat_features:]
        
        colors = cm.rainbow(np.linspace(0, 1, y.shape[1]))
        fig = plt.figure()
        for i in range(x.shape[1]):
            plt.plot(x[:,i], y[:,i], marker='none', linestyle='-', linewidth=1, color=colors[i])
        plt.xlabel('Standard Deviation of Noise')
        plt.ylabel('Avg. MSEs (per obs.) over whole train set (cont. features)')
        fig.savefig(os.path.join(logdir, prefix + 'timewarp_cont_loss.png'), dpi=300, bbox_inches='tight')
        plt.close()
        

    

def plot_feature_distributions(logdir, X_cat_train, X_cont_train, y_train, 
                               X_cat_gen, X_cont_gen, y_gen, task):
    
    for i in range(X_cat_gen.shape[1]):
        
            counts_true = pd.DataFrame(X_cat_train[:,i]).value_counts()
            props_true = counts_true / sum(counts_true)
            counts_gen = pd.DataFrame(X_cat_gen[:,i]).value_counts()
            props_gen = counts_gen / sum(counts_gen)
            
            x = []
            y_true_list = []
            y_gen_list = []
            
            for cat in props_true.keys():
                x.append(cat[0])
                y_true_list.append(props_true[cat])
                if cat in props_gen.keys():
                    y_gen_list.append(props_gen[cat])
                else:
                    y_gen_list.append(0.0)
 
            fig, ax = plt.subplots(1, 1)
            ax.bar(np.arange(len(x)) - 0.1, y_true_list, color='tab:orange', width=0.2, label='true')
            ax.bar(np.arange(len(x)) + 0.1, y_gen_list, color='tab:blue', width=0.2, label='generated')
            ax.set_xticks(np.arange(len(x)), x)
            ax.set_xticklabels(x)
            plt.legend()
            
            if i == 0 and task != 'regression' and y_train is None:
                label = 'dist_target.png'
            else:
                label = f'dist_cat_feature_{i}.png'
            
            fig.savefig(os.path.join(logdir, 'data', label), dpi=300)
            plt.close()
            
    for i in range(X_cont_gen.shape[1]):
        fig, ax = plt.subplots(1, 1)
        pd.DataFrame(X_cont_train[:,i]).plot(kind='density', ax=ax, color = 'tab:orange')
        pd.DataFrame(X_cont_gen[:,i]).plot(kind='density', ax=ax, color = 'tab:blue')
        plt.legend(['true', 'generated'])
        plt.axvline(x = X_cont_train[:,i].mean().item(), color = 'tab:orange')
        plt.axvline(x = X_cont_gen[:,i].mean().item(), color = 'tab:blue')
        
        if task == 'regression':
            label = f'dist_cont_feature_{i}.png' if i > 0 else 'dist_target.png'
        else:
            label = f'dist_cont_feature_{i}.png'

        fig.savefig(os.path.join(logdir, 'data', label), dpi=300)
        plt.close()
        
    if y_train is not None:
        
        counts_true = pd.DataFrame(y_train).value_counts()
        props_true = counts_true / sum(counts_true)
        counts_gen = pd.DataFrame(y_gen).value_counts()
        props_gen = counts_gen / sum(counts_gen)
        
        x = []
        y_true_list = []
        y_gen_list = []
        
        for cat in props_true.keys():
            x.append(cat[0])
            y_true_list.append(props_true[cat])
            if cat in props_gen.keys():
                y_gen_list.append(props_gen[cat])
            else:
                y_gen_list.append(0.0)

        fig, ax = plt.subplots(1, 1)
        ax.bar(np.arange(len(x)) - 0.1, y_true_list, color='tab:orange', width=0.2, label='true')
        ax.bar(np.arange(len(x)) + 0.1, y_gen_list, color='tab:blue', width=0.2, label='generated')
        ax.set_xticks(np.arange(len(x)), x)
        ax.set_xticklabels(x)
        plt.legend()
        
        fig.savefig(os.path.join(logdir, 'data', 'dist_target.png'), dpi=300)
        plt.close()
        
        
def plot_joint_density(logdir, xvar, yvar, xmin, xmax, ymin, ymax, label, xlabel, ylabel):
    
    ax = sns.jointplot(x=xvar, y=yvar, kind='kde', color="tab:blue", space=0,
                  xlim=(xmin, xmax), ylim=(ymin, ymax))
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    ax.fig.subplots_adjust(bottom=.1, left=.1)
    plt.savefig(os.path.join(logdir, label + '.png'), dpi=300, bbox_inches = "tight")
    plt.close()
    
    

def get_additional_adult_plots(logdir, X_cont_train, X_cont_gen):
        
    x1_real = X_cont_train[:,1][X_cont_train[:,1] > 0]
    x1_zeros_real = (X_cont_train[:,1] == 0).sum()
    x1_gen = X_cont_gen[:,1][X_cont_gen[:,1] > 0]
    x1_zeros_gen = (X_cont_gen[:,1] == 0).sum()
    n = X_cont_train.shape[0]
    
    f, ax = plt.subplots()
    plt.hist(x1_real, bins = 100, alpha = 0.5, label = 'real')   # blue
    plt.hist(x1_gen, bins = 100, alpha = 0.3, label = 'generated') # orange
    plt.text(.01, .99, 
        f'# real zeros = {x1_zeros_real}, # gen zeros = {x1_zeros_gen}, n = {n}', 
        ha='left', va='top', transform=ax.transAxes)
    plt.xlabel('Capital Gain')
    plt.ylabel('Samples')
    plt.savefig(os.path.join(logdir, 'capital_gain.png'), dpi=300, bbox_inches = "tight")
    plt.close()
    
    # capital loss
    x2_real = X_cont_train[:,2][X_cont_train[:,2] > 0]
    x2_zeros_real = (X_cont_train[:,2] == 0).sum()
    x2_gen = X_cont_gen[:,2][X_cont_gen[:,2] > 0]
    x2_zeros_gen = (X_cont_gen[:,2] == 0).sum()

    f, ax = plt.subplots()
    plt.hist(x2_real, bins = 100, alpha = 0.5, label = 'real')  # blue
    plt.hist(x2_gen, bins = 100, alpha = 0.3, label = 'generated') # orange
    plt.text(.01, .99, 
        f'# real zeros = {x2_zeros_real}, # gen zeros = {x2_zeros_gen}, n = {n}', 
        ha='left', va='top', transform=ax.transAxes)
    plt.xlabel('Capital Loss')
    plt.ylabel('Samples')
    plt.savefig(os.path.join(logdir, 'capital_loss.png'), dpi=300, bbox_inches = "tight")
    plt.close()

    plot_joint_density(logdir, 
                        X_cont_train[:,3], 
                        X_cont_train[:,4], 
                        0, 85, 0, 17,
                        label='age_educ_joint_density_true', 
                        xlabel = 'age', 
                        ylabel = 'education (num)')
    
    plot_joint_density(logdir, 
                        X_cont_gen[:,3], 
                        X_cont_gen[:,4], 
                        0, 85, 0, 17,
                        label='age_educ_joint_density_generated', 
                        xlabel = 'age', 
                        ylabel = 'education (num)')
    
